{{
***************************************************
* File jvmObject: miscellaneous object operations *
***************************************************
}}

{{
  Java manipulates two types of fields: dynamic and static. Whereas dynamic fields change, static fields do 
  not. These fields are manipulated using simple get and put mechanisms.  

  Putting Fields. 
  The put mechanism uses byte1 and byte2 to construct an index into the constant pool of the current 
  class. The item at the index is a field reference to a class name and a field name, which is resolved to a 
  field block pointer that has both the field width and the field offset. The field at the offset from the start of 
  the object referenced is set to the value on the top of the stack. 
  Note:
  A NullPointerException is generated if the referenced object is null.
  For static fields, an IncompatibleClassChangeError is thrown if the specified field is a static field.

  Getting Fields. 
  The get mechanism uses byte1 and byte2 to construct an index into the constant pool of the current 
  class. The item at the index is a field reference to a class name and a field name, which is resolved to a 
  field block pointer that has both the field width and the field offset. The value at the offset replaces the 
  object reference on the top of the stack. 
  Note:
  A NullPointerException is generated if the referenced object is null.
  For static fields, an IncompatibleClassChangeError is thrown if the specified field is a static field. 
}}
{{
  JEM object/array storage.
  Objects are stored on the heap in the following format:
    int class          'offset to class header
    byte data[length]  'the space occupied by the object data
  Note that single boolean, byte and char variables each occupy 2 bytes, just as short, int and reference.
  For arrays only the references to the arrays are stored in the object. The arrays themselves are objects
  on the heap.
}}


OBJ
  js: "jvmStack"
  jc: "jvmConvert"
  jd: "jvmData"
  jh: "jvmHeap"
  
  
PUB jb_putfield | aref,value,offset
{{ Meaning: Set 32-bit field in object. }}
{{
  JEM format: putfield,byte1,byte2
  byte1 and byte2 form 16bits offset into object instance on heap: offset = (byte1<<8)|(byte2) 
}}
  value := js.pop 'get value
  aref := js.pop  'get reference
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  jd.writeJavaIntBE(aref + offset,value)
  jd.incJvmPC(3)
  

PUB jb_putstatic | value,offset
{{ Meaning: Set 32-bit static field in class. }}
{{
  JEM format: putstatic,byte1,byte2
  byte1 and byte2 form 16bits offset into static variables area: offset = (byte1<<8)|(byte2) 
}}
  value := js.pop 'get value
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  jd.putStatic(offset,value)
  jd.incJvmPC(3)
  

PUB jb_getfield | aref,value,offset
{{ Meaning: Fetch 32-bit field from object. }}
{{
  JEM format: getfield,byte1,byte2
  byte1 and byte2 form 16bits offset into object instance on heap: offset = (byte1<<8)|(byte2) 
}}
  aref := js.pop  'get reference
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  value := jd.readJavaIntBE(aref + offset)
  js.push(value) 'return value
  jd.incJvmPC(3)


PUB jb_getstatic | value,offset
{{ Meaning: Get 32-bit static field from class. }}
{{
  JEM format: getstatic,byte1,byte2
  byte1 and byte2 form 16bits offset into static variables area: offset = (byte1<<8)|(byte2) 
}}
  offset := jd.readJavaIntBE(jd.getJvmPC+1)
  value := jd.getStatic(offset)
  js.push(value) 'return value
  jd.incJvmPC(3)


PUB jb_new | i,size,class,aref
{{ Meaning: Create new object. }}
{{
  Description: byte1 and byte2 are used to construct an index into the constant pool of the 
  current class, which must be a class name that can be resolved to a class pointer. A new 
  instance of that class is then created, and a reference to the object is pushed on the stack.
}}
{{
  JEM format: new,byte1,byte2
  byte1 and byte2 form 16bits offset to class descriptor: class = (byte1<<8)|(byte2) 
}}
  class := jd.readJavaIntBE(jd.getJvmPC+1)
  size := jd.readJavaByte(class+2) 'get size of non-static non-final variables
  if size > 0
    aref := jh.allocate(size + 2)
    if aref <> 0
      jd.writeJavaIntBE(aref,class)
      i := 2
      repeat size 'clear databytes
        jd.writeJavaByte(aref+i,$00)
        i++
      js.push(aref) 'push reference
      jd.incJvmPC(3)
    else
      jd.setException
      js.push($0000) 'do we need return value ??
      jd.incJvmPC(3)
  else
    jd.setException
    js.push($0000)  'do we need return value ??
    jd.incJvmPC(3)
    

PUB jb_checkcast
{{ Meaning: Make sure object is of given type. }}
{{
  Description: Determines whether the referenced object can be cast to be a reference to an 
  object of another class. A null object reference can be cast to any class. Otherwise, the 
  referenced object must be an instance of the expected class or one of its superclasses. byte1  
  and byte2 are used to construct an index into the constant pool of the current class, which is 
  presumed to be a class name that can be resolved to a class pointer. 
  Note:
  A ClassCastException is thrown if the referenced object cannot be cast to the expected class.
}}
{{
  JEM format: checkcast,byte1,byte2
  byte1 and byte2 are always 0 (so this bytecode does not do anything)
}}
  jd.incJvmPC(3)
  

PUB jb_instanceof | instance,class,instanceclass
{{ Meaning: Determine if an object is of given type and return result. }}
{{
  Description: Determines whether the referenced object can be cast to be a reference to an 
  object of the expected class. This instruction will overwrite the object reference with 1 if the 
  object is an instance of the expected class or one of its superclasses. Otherwise, the object 
  reference is overwritten by 0.
}}
{{
  JEM format: instanceof,byte1,byte2
  byte1 and byte2 form 16bits offset to class header: offset = (byte1<<8)|(byte2) 
}}
  instance := js.pop 'get instance reference
  class := jd.readJavaIntBE(jd.getJvmPC+1) 'class type
  instanceclass := jd.readJavaIntBE(instance) 'get class type for instance
  repeat
    if instanceclass == class
      js.push(1)
      jd.incJvmPC(3)
      return
    else
      class := jd.readJavaIntBE(class) 'get superclass
  until class == $FFFF
  js.push(0)
  jd.incJvmPC(3)
  
  